/*---------------------------------------------------------------------------
File Name:      QP_MemRedCost.h
Author:         Alexandre Belloni
Last Modified:  05-OCT-05

Description:    This file has a derived class from TQPclass. In our applications
				the storage of the matrix Q and the computation of the reduced
                costs are intrinsically connected.
                Thus, storage of Q and computation of Reduced Costs are
                performed together to provide enough flexibility/efficiency
                in our implementation.

				We still assume that a routine to compute an element Qij is
                provided (via a virtual function here).
---------------------------------------------------------------------------*/
//---------------------------------------------------------------------------

#ifndef QP_MemRedCostPathH
#define QP_MemRedCostPathH
//---------------------------------------------------------------------------
#include "BasicInit.h"
#include "QPclassPath.h"

#define MemStyleQ       2
//    Value     Implementation
//      0       no memory allocation at all. All computations on the fly for Qij.
//      1       only columns of Q associated with Is are stored.
//      2       complete columns of Qs are stored but values are computed only if needed.
//      3       Only Qss and Qsw where w = { Io' U Ic'} are stored

#define ReducedCostMode 3
//   Value      Implementation
//      0       Index with minimum reduced cost is inserted on every iteration
//      1       Using Io', Io'', Ic', Ic''... minimum reduced cost within Io' and Ic'
//      2       First index with reduce cost less than RedCostBar within Io' and Ic'.
//      3       Work with Iopp, Icpp for Major Iterations, Icp and Iop for small iterations



class TQPMemRedCostPath : public TQPclassPath {

//---------------------------------------------------------------------------
// Interface with the user should be only via these routines
//---------------------------------------------------------------------------
public:

        TQPMemRedCostPath();
        ~TQPMemRedCostPath();
        void TQPMemRedCostPathInit( Index Max_n, Index n, hpVec Y0, hpVec V0, hpVec C0 );

        void            SetTheta( hpReal val );
        void            SolvePath ( hpReal InitVal, hpReal FinalVal );
        void            SolveApproximatePath (hpReal InitVal, hpReal FinalVal );
		void            SolveQP ( hpReal Parameter );
		inline Index	GetCurrentPathMode();
		inline Index    GetNumBreakPoints();
		inline Index    GetNumJumps();
		inline Index    GetCountMajorLoop();
		inline Index    GetCountCallsReducedCost();
		inline void		SetDataTestMode( IntNumber val );
		boolean         EvaluatingTestData();
		inline IntNumber GetDataTestMode();
		inline hpReal   GetInitParameter();
		inline hpReal   GetFinalParameter();
		
protected:
		virtual void    UpdateTestDataIc( Index i, TypeEvent Event ) = 0;
		virtual void    InitTestDataIc() = 0;

		void 			ReturnPathToOpt ();
		inline hpReal   GetRedCostBar();
		inline Index    GetCurrentAllocQ();
		inline Index    GetNcp();
		inline Index    GetNop();
		inline Index    GetNq();
		inline hpReal   GetQicCc ( Index i );
		inline Index    GetNumberOfTies();
		inline Index    GetNumberOfTiesStep2();
		inline TypeEvent GetParametricEventStep2();
		inline Index    GetParametricIndexStep2();
		inline hpReal   Get_f_deriv();
		void            Compute_f_deriv();

		inline void		SetMaxToJump( IntNumber val );
		inline void		SetMaxSizeNoPath( IntNumber val );
		inline void		SetMaxSizeNcPath( IntNumber val );
		inline void		SetBPtoUpdates( IntNumber val );

		inline void		SetMaxSizeNop( IntNumber  val );
		inline void		SetMaxSizeNcp( IntNumber  val );
		inline void		SetCountToIopp( IntNumber val );
		inline void		SetCountToIcpp( IntNumber  val );
		inline void		SetCountToIoppp( IntNumber val );
		inline void		SetCountToIcppp( IntNumber val );
		inline void		SetBlockAllocQ ( IntNumber val );
		inline void		SetMaxNumberofQcolumns ( IntNumber val );
		inline void		SetBreakPointIntervals ( IntNumber val );
		inline void		SetNumberOfTests ( IntNumber val );
		inline void             SetPathMode( IntNumber val );
                inline void             SetInitParameter( hpReal val );
                inline void             SetFinalParameter( hpReal val );

		#if ( TimeMode )

		hpReal          TimeKKTsystemQP;
		hpReal          TimeReduceCostQP;
		hpReal          TimeAddToBaseQP;
		hpReal          TimeRemoveFromBaseQP;
		hpReal          TimeInitQP;
        hpReal          TimeTotalAllQP;

        hpReal          TimeBreakStep1;
        hpReal          TimeBreakStep2;
        hpReal          TimeUpdates;
        hpReal          TimeCorrector;
		void            InitTimePath ();

        #endif


private:

//----------------------------------------------------------------------------
//      Computation of Regularized Path
//----------------------------------------------------------------------------
        void            PathMinRatioTestStep2LowerBound( hpReal Bound );
        void            PathMinRatioTestStep2( hpReal Bound );
        boolean         GetNextBreakPoint ();
		inline hpReal   GetCurrentMinRationStep2Pos();
		inline hpReal   GetCurrentMinRationStep2Neg();
		void            UpdatePathVariable();
		void            InitStep2Path ();
        boolean         GetApproxBreakPoint ();
		boolean         CheckApproximateStatus();
		void            InitApproxStep2Path();
        void            UpdateApproxPathVariable();
        void            PathMinRatioTestApproxStep2( hpReal Bound );
//----------------------------------------------------------------------------
//      Reduced Cost for the Standard QP
//----------------------------------------------------------------------------

        inline void     UpdateReducedCostBar ();
        boolean         NeedToGetNewIndex ();
		boolean         MajorIterationsReducedCosts( Index &place_k, Index &IndexMin,
										hpReal &RCMin, TypeEvent &NextEvent );
		void            RedCostIndexAddedToIc ( Index i );
        void            RedCostIndexAddedToIo ( Index i );
        void            RedCostIndexRemovedFromIc ( Index i );
		void            RedCostIndexRemovedFromIo ( Index i );

		QPErrorMsg 		CheckingParametricStep();

		// Interface with "real" data set...
        virtual hpReal  GetDataQij( Index i, Index j ) = 0;     // Compute only 1 element
        virtual void    GetDataQi ( hpVec Qi, Index i) = 0;     // Compute Complete Row
        virtual hpReal  GetMaxQi( Index i ) = 0;                // return a bound on max | Q_ij |
        virtual boolean PositiveKernel() = 0;                // return True if the Kernel is positive


        // This allows for different squemes for adding variables
        inline hpReal   GetQij ( Index i, Index j );
                                // return element Qij
        inline hpReal   SprodQi_dIset ( Index i, hpVec dIset, IndexVec Iset, Index Nset);
                                // return < Qi[Iset], dIset[1...Nset] >, NOTE: dIset has size Nset
        inline hpReal   SprodQi_d ( Index i, hpVec d, IndexVec Iset, Index Nset);
                                // return < Q[Iset], d[Iset] >, NOTE: d has size n

		// Possible Interface...
		virtual void    UpdateEveryIteration () = 0;
        virtual void    UpdateEveryIterationPath () = 0;

        // Test Data
		virtual hpReal  ComputingPerformanceOnTestData()=0;
//---------------------------------------------------------------------------
//      Regularized Path
//---------------------------------------------------------------------------
		Index 			PathMode; 	//      0       All reduced costs are computed exactly before moving to the next breakpoint
									//      1       LowerBound of the reduced costs are kept. True reduced cost is computed only if necessary
									//      2       Approximate Solution...
		boolean			CheckedPath;
		boolean         ParametricMode;         // True if running Parametric Mode.
        QPErrorMsg      ParametricStatus;
		TypeEvent       ParametricEvent;
		TypeEvent       ParametricEventStep2;
		IntNumber 		CountToJump;			// count the number of sequential breakpoints without progress
		IntNumber		MaxToJump;				// Number of breakpoints to jump
		Index           NumJumps;         		// Number of Jumps made so far
		Index           NumBreakPoints;         // Number of Breakpoints found so far
		hpVec           LowerReducedCost;       // Constains a Lower bound for the reduced costs
        hpVec           LowerSlope;             // Contains < Qis, alpha_s >
		hpVec           QisAlpha;               // Contains < Qis, d_s >
		Index           NumberOfTies;           // Number of Ties
        Index           NumberOfTiesStep2;      // Number of Ties of Reduced Cost
        Index           ParametricIndex;        // Corresponding binding index
        Index           ParametricIndexStep2;   // Corresponding binding index
        hpReal          MinRatioStep2Pos;
        hpReal          MinRatioStep2Neg;
		hpReal          NextTheta;
		hpReal          OldTheta;
        hpReal          InitParameter;
        hpReal          FinalParameter;
		hpReal          f_deriv;

        Index           NoPath;           //
        Index           NcPath;           //
        IndexVec        IoPath;           // Subset of Io to be searched for reduced costs in the end
        IndexVec        IcPath;           // Subset of Ic to be searched for reduced costs in the end
        Index           MaxSizeNoPath;    // Maximum Size of NoPath
        Index           MaxSizeNcPath;    // Maximum Size of NcPath
        Index           BPtoUpdates;      // Number of Break Points before Updating

//---------------------------------------------------------------------------
//      Testing Solution on Training Data
//---------------------------------------------------------------------------
		hpReal BestTestValue;
		hpReal BestThetaTestValue;

		hpReal LastThetaTested;
		hpReal NumberOfTests;

		Index  BreakPointIntervals;

		hpReal quantile_current;
		hpReal tau_value;
		hpReal DiffAlpha;//ComputeDiffAlpha();

		IntNumber DataTestMode;


//---------------------------------------------------------------------------
//      Working data...
//---------------------------------------------------------------------------
        hpVec           tmpVecM;        // working vector...
		IndexVec        tmpIndVec;
//---------------------------------------------------------------------------
//  Memory Allocation for Q....
//---------------------------------------------------------------------------
        QPErrorMsg      AddColumnToQ( Index i );        // Used only if MemStyle == 1 or 2
        QPErrorMsg      FreeMemQ( Index i );
        void            DeleteQ();
        QPErrorMsg      VerifyQPMem();
//---------------------------------------------------------------------------
// Flags and Indicators
//---------------------------------------------------------------------------

hpReal          RedCostBar;             // interup the search if RedCost < RedCostBar
Index           Nop;                    // Nop = |Iop|
Index           Ncp;                    // Ncp = |Icp|
IndexVec        Iop;                    // Subset of Io to be searched for reduced costs first
IndexVec        Icp;                    // Subset of Ic to be searched for reduced costs first
IntVec          CounterNonNegRedCost;   // Number of Consecutive Iterations with positive reduced costs
                                        // on the current set it belongs...
IntNumber       CountToIopp;            // After CountToIopp Index is Removed from Iop
IntNumber       CountToIcpp;            // After CountToIcpp Index is Removed from Icp

hpVec           QC_redcost;             // ith component stores QicCc.

//#if     ( ReducedCostMode == 3 )
		boolean         MajorLoop;      // 1 if we can call Major Loop, 0 otherwise
		Index           Nopp;           // Nop = |Iopp|
		Index           Ncpp;           // Ncp = |Icpp|
		IndexVec        Iopp;           // Subset of Io to be searched for reduced costs in the end
		IndexVec        Icpp;           // Subset of Ic to be searched for reduced costs in the end
		Index           MaxSizeNop;     // Maximum Size of Nop
		Index           MaxSizeNcp;     // Maximum Size of Ncp
		IntNumber       CountToIoppp;
		IntNumber       CountToIcppp;
//#endif
		Index           CountCallsReducedCost;
        Index           CountMajorLoop; // counts the number of major loops

//----------------------------------------------------------------------------
// Memmory Allocation Parameters
//----------------------------------------------------------------------------

IntNumber       BlockAllocQ;            // Number of Rows allocatted at once (we must have BlockAlloc < Max_n)
IntNumber       CurrentAllocQ;          // Current Number of allocated rows
IntNumber       MaxNumberofQcolumns;    // Maximum number of columns that the user allow us to store.
										// CurrentAlloc <= MaxNumberofQcolumns

hpMat           QQ;                     // Parcial Q Matrix used for some memory representations
IndexVec        QIndex;                 // QIndex[ i ] = j,     if the ith column of Q is stored at QQ[j]
                                        //                      j <= MaxNumberofQcolumns
                                        //             = Max_n, if ith column is not stored
IndexVec        InvQIndex;              // InvQIndex[ j ] = i => QQ[j] contains information about ith row of Q

Index           Nq;                     // Number of columns

//----------------------------------------------------------------------------
// Output
//-----------------------------------------------------------------------------
void            PrintConfigurationInfo ();
}; //----- end of class ----


//-----------------------------------------------------------------------------
// Inline functions
//-----------------------------------------------------------------------------

inline hpReal   TQPMemRedCostPath::GetQij ( Index i, Index j )
{

#if     ( MemStyleQ == 0 )

        return GetDataQij(i,j);

#elif   ( MemStyleQ == 1 )

        if ( QIndex[ i ] < Index ( CurrentAllocQ ) ) {
                return QQ[ QIndex[ i ] ][ j ];
				}
        else if ( QIndex[ j ] < Index ( CurrentAllocQ ) ) {
                return QQ[ QIndex[ j ] ][ i ];
                }
		else return GetDataQij(i,j);

#elif   ( MemStyleQ == 2 )
		//Index tmp;
        //for ( long k = 1 ; k < 1 ; k++ )
		//        tmp = floor(sin(k));
        register Index QIndi = QIndex[ i ];
        register Index QIndj = QIndex[ j ];
        if ( QIndi < Index ( CurrentAllocQ ) ) {
                if ( QQ[ QIndi ][ j ] < hpInf ){
                        return QQ[ QIndi ][ j ];
                        }
                else    {
                        return QQ[ QIndi ][ j ] = GetDataQij(i,j);
                        }
                }
		else if ( QIndj < Index ( CurrentAllocQ ) ) {
                if ( QQ[ QIndj ][ i ] < hpInf ){
                        return QQ[ QIndj ][ i ];
                        }
                else    {
                        return QQ[ QIndj ][ i ] = GetDataQij(i,j);
                        }
                }
        else return GetDataQij(i,j);


#elif ( MemStyleQ == 3 )

        PrintMsgError("MemStyleQ == 3 not available for this version\n Change MemStyle.");

#endif

}
//-----------------------------------------------------------------------------
inline hpReal   TQPMemRedCostPath::SprodQi_dIset ( Index i, hpVec dIset, IndexVec Iset, Index Nset) // uses Qi[Iset] dIset[1...Nset] has sive Nset
{
register hpReal t = hpZero;
for (register Index k = Nset ; k-- ; )
		t += dIset[k]*GetQij(Iset[k],i);
return t;
}

inline hpReal   TQPMemRedCostPath::SprodQi_d ( Index i, hpVec d, IndexVec Iset, Index Nset)  // uses Q[Iset] d[Iset]
{
register hpReal t = 0;
for (register Index k = Nset ; k-- ; )
		t += d[Iset[k]]*GetQij(Iset[k],i);
return t;
}

//-----------------------------------------------------------------------------


#if ( TimeMode && ReducedCostMode == 2 )

inline void TQPMemRedCostPath::UpdateReducedCostBar ()
{
// As a rule of thumb, we want the search for the reduce cost to be comparable
// with updating the data structures (timewise)

if ( TimeReduceCost > GetAbsTolError() ){
		hpReal a = (TimeKKTsystem + TimeAddToBase + TimeRemoveFromBase) / TimeReduceCost ;
		RedCostBar = GetMinReducedCost()*a;//- sqrt( abs(RedCostBar) * a );//*= sqrt(a);// - max( 0.1, min( a , 10.0 ) );
		}
else if ( (TimeKKTsystem + TimeAddToBase + TimeRemoveFromBase) > GetAbsTolError() ){
		RedCostBar *= 2;
		}
}

#endif


//-----------------------------------------------------------------------------
// In line functions
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetMaxSizeNoPath( IntNumber val )
{
MaxSizeNoPath = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetMaxSizeNcPath( IntNumber val )
{
MaxSizeNcPath = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetBPtoUpdates( IntNumber val )
{
BPtoUpdates = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetBreakPointIntervals ( IntNumber val )
{
BreakPointIntervals = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetNumberOfTests ( IntNumber val )
{
NumberOfTests = hpReal ( val );
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetBlockAllocQ ( IntNumber val )
{
BlockAllocQ = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetMaxNumberofQcolumns ( IntNumber val )
{
MaxNumberofQcolumns = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetMaxToJump( IntNumber val )
{
MaxToJump = val;
}
//----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetMaxSizeNop( IntNumber  val )
{
MaxSizeNop = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetMaxSizeNcp( IntNumber  val )
{
MaxSizeNcp = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetCountToIopp( IntNumber val )
{
CountToIopp = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetCountToIcpp( IntNumber val )
{
CountToIcpp = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetCountToIoppp( IntNumber val )
{
CountToIoppp = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetCountToIcppp( IntNumber val )
{
CountToIcppp = val;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPMemRedCostPath::GetQicCc ( Index i )
{
return QC_redcost[i];
}
//-----------------------------------------------------------------------------
inline Index    TQPMemRedCostPath::GetNq()
{
return Nq;
}
//-----------------------------------------------------------------------------
inline Index    TQPMemRedCostPath::GetNop()
{
return Nop;
}
//-----------------------------------------------------------------------------
inline Index    TQPMemRedCostPath::GetNcp()
{
return Ncp;
}
//-----------------------------------------------------------------------------
inline Index    TQPMemRedCostPath::GetCurrentAllocQ()
{
return CurrentAllocQ;
}
//-----------------------------------------------------------------------------
inline hpReal    TQPMemRedCostPath::GetRedCostBar()
{
return RedCostBar;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPMemRedCostPath::GetCurrentMinRationStep2Pos()
{
return MinRatioStep2Pos;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPMemRedCostPath::GetCurrentMinRationStep2Neg()
{
return MinRatioStep2Neg;
}
//-----------------------------------------------------------------------------
inline Index  TQPMemRedCostPath::GetNumberOfTiesStep2()
{
return NumberOfTiesStep2;
}
//-----------------------------------------------------------------------------
inline Index  TQPMemRedCostPath::GetNumberOfTies()
{
return NumberOfTies;
}
//-----------------------------------------------------------------------------
inline TypeEvent  TQPMemRedCostPath::GetParametricEventStep2()
{
return ParametricEventStep2;
}
//-----------------------------------------------------------------------------
inline Index  TQPMemRedCostPath::GetParametricIndexStep2()
{
return ParametricIndexStep2;
}
//-----------------------------------------------------------------------------
inline hpReal TQPMemRedCostPath::GetInitParameter()
{
return InitParameter;
}
//-----------------------------------------------------------------------------
inline hpReal TQPMemRedCostPath::GetFinalParameter()
{
return FinalParameter;
}
//-----------------------------------------------------------------------------
inline void TQPMemRedCostPath::SetInitParameter( hpReal val )
{
InitParameter = val;
}
//-----------------------------------------------------------------------------
inline void TQPMemRedCostPath::SetFinalParameter( hpReal val )
{
FinalParameter = val;
}
//-----------------------------------------------------------------------------
inline Index  TQPMemRedCostPath::GetNumBreakPoints()
{
return NumBreakPoints;
}
//-----------------------------------------------------------------------------
inline Index  TQPMemRedCostPath::GetNumJumps()
{
return NumJumps;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPMemRedCostPath::Get_f_deriv()
{
return f_deriv;
}
//-----------------------------------------------------------------------------
inline Index   TQPMemRedCostPath::GetCountMajorLoop()
{
return CountMajorLoop;
}
//-----------------------------------------------------------------------------
inline Index   TQPMemRedCostPath::GetCountCallsReducedCost()
{
return CountCallsReducedCost;
}
//-----------------------------------------------------------------------------
inline Index   TQPMemRedCostPath::GetCurrentPathMode()
{
return PathMode;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetPathMode( IntNumber val )
{
PathMode = val;
}
//-----------------------------------------------------------------------------
inline void		TQPMemRedCostPath::SetDataTestMode( IntNumber val )
{
if ( val >= 0 && val <= 2 ) {
	DataTestMode = val;
	}
else {
	DataTestMode = -1;
	}
}
//-----------------------------------------------------------------------------
inline IntNumber	TQPMemRedCostPath::GetDataTestMode()
{
return DataTestMode;
}

#endif





