/*---------------------------------------------------------------------------
File Name:      QPclassPath.h
Author:         Alexandre Belloni
Last Modified:  05-OCT-05

Description:    This file has the basic class to solve QPs of the form:

                min 1/2 < alpha, Q alpha > - < v, alpha >
                 s.t.              < y, alpha >  = 0
                              0<=       alpha   <= c

                where Q is sdp.

                Requires interface with Q and reduce cost computation.
---------------------------------------------------------------------------*/
#ifndef QPclassPathH
#define QPclassPathH
//---------------------------------------------------------------------------
#include "BasicInit.h"
#include "VectorOperations.h"
#include "MatrixOperations.h"
//---------------------------------------------------------------------------
// Definitions
//---------------------------------------------------------------------------

#define DebugMode       0
//    Value     Implementation
//      0       no data check are performed.
//      1       r1, r2, lin_obj, Io, Ic, Is, YcCc, cte_obj, and Optimality are checked in the end
//      2       as in DebugMode == 1 but on every outer loop
//      3       as in DebugMode == 1 but on every outer & inner loop
//      4       as in DebugMode == 3 plus L and Q.

#define ExactMode       1
//    Value     Implementation
//      0       <r1,r1>, <r2,r2>, and <r1,r2> are not recomputed
//      1       <r1,r1>, <r2,r2>, and <r1,r2> are recomputed whenever needed.

#define L_LowerTri      1
//   Value      Implementation
//      0       L is a full matrix Ns by Ns... easier to exchange lines
//      1       L is allocated to be lower triangular.

//#define PrintMode       1
//    Value     Implmentation
//      0       No Print
//      1       Simple OutPut
//      2       Detailed Output  (every iteration)

//---------------------------------------------------------------------------
// Types
//---------------------------------------------------------------------------


typedef enum{
              NoEvent           =  0,
              Io2Is             =  1,
              Is2Ic             =  2,
              Is2Io             =  3,
              Ic2Is             =  4

        } TypeEvent;

typedef enum{
              No_Set            = -1,
              Set_Zero          = 0 ,
              Set_Active        = 1 ,
              Set_C             = 2

        } TypeSets;

typedef enum{
              MsgOk             = 0,
              MsgNumericalProb  = 2,
              MsgMemory         = 4,
              MsgInvalidIndex   = 6,
              MsgQPIncorrect    = 8,
              MsgFatal          = 10,
			  MsgFileError      = 12,
			  MsgPotentialCycle = 14,
			  MsgCycle          = 16,
			  MsgOptimalityLost = 18,
			  MsgReoptimized	= 20
		} QPErrorMsg;

//---------------------------------------------------------------------------

class TQPclassPath {

//---------------------------------------------------------------------------
// Interface with the user should be only via these routines
//---------------------------------------------------------------------------
public:

        TQPclassPath();
        ~TQPclassPath();
        QPErrorMsg      TQPclassPathInit( Index Max_n0, Index n0, hpVec Y0, hpVec V0, hpVec C0 );
		QPErrorMsg      SolveQP();

		void            SetNoBias( boolean val );

		hpReal          ReadOptimalValue();
		void            ComputeObjFuncFast();
		//		inline hpReal   GetTotalTime ();
		//		inline hpReal   GetQPTime ();

		inline void     SetC ( hpVec C0 ) ;             // can change C only before solving
		inline void     SetC ( hpReal C0 );             // can change C only before solving

		inline void     SetAbsPrecision( hpReal val );
		inline void     SetRelPrecision( hpReal val );

		inline IntNumber GetIteration();
		inline IntNumber GetInnerIter();
		inline void      SetMaxIteration ( IntNumber val );

		inline hpReal   Getf_obj();
		inline hpReal   Getb();
		inline hpReal   Getb_deriv();

		inline Index    Getn();
		inline Index    GetDegreeDependence();
		inline hpReal   GetAlpha( Index k );
		inline Index    GetNs();
		inline Index    GetNo();
		inline Index    GetNc();
		inline Index    GetIs( Index k );
		inline Index    GetIo( Index k );
		inline Index    GetIc( Index k );
		inline void     CopyCurrentIs( IndexVec Vec );
		inline void     CopyCurrentIo( IndexVec Vec );
		inline void     CopyCurrentIc( IndexVec Vec );
		void            CopyCurrentPointSparse( hpVec Vec );
		void            PrintCurrentSolution ( FILE *out );
		void            PrintCurrentSolutionSparse ( FILE *out );

protected:
		inline bool 	BelongToIo ( Index k );
		inline bool 	BelongToIc ( Index k );
		inline bool 	BelongToIs ( Index k );

		inline Index    GetMax_n();
		inline hpReal   GetYi(Index i);
		inline hpReal   GetVi(Index i);
		inline hpReal   GetCi(Index i);
		inline hpReal   Getr1r1 ();
		inline hpReal   Getr1z1 ();

		inline hpReal   GetAbsTolError ();
		inline hpReal   GetRelTolError ();

		inline hpReal   GetStepSize();
		inline hpReal   GetMinReducedCost();
		inline Index    GetNewIndexToBeAdded();
		inline Index    GetCurrentAllocL();

		inline IntNumber GetCase0Count();
		inline IntNumber GetCase1Count();
		inline IntNumber GetCase2Count();
		inline IntNumber GetInfiniteDirectionCount();

		inline void     SetNewIndexToBeAdded ( Index i , hpReal val );
		inline void     SetBaseEvent ( TypeEvent CurrentEvent );
		inline void 	SetBlockAllocL ( IntNumber val );
		QPErrorMsg 		JumpSolveQP( hpReal JumpTheta );
//----------------------------------------------------------------------------
//  Time Variables...
//----------------------------------------------------------------------------
		#if ( TimeMode )
		struct timeb    TimeStartCount;
		struct timeb    TimeFinishCount;
		hpReal          TimeKKTsystem;
		hpReal          TimeReduceCost;
		hpReal          TimeAddToBase;
		hpReal          TimeRemoveFromBase;
		hpReal          TimeInit;
		hpReal          TimeTotalAll;
		struct timeb    TimeStartAll;

		inline void     TimeStart();
		inline void     TimeFinish();
		inline void     TimeAdd( hpReal *time_total );
		inline void     TimeTotal();

		#endif
//----------------------------------------------------------------------------
// Output
//-----------------------------------------------------------------------------
		void            PrintBasicInfo ();
		void            PrintErrorMsg ( char *s );
		void            PrintInformativeMsg ( char *s );
		virtual void    PrintConfigurationInfo ()=0;
		inline	void	SetPrintMode( IntNumber val );

		QPErrorMsg      SolveCurrentBase ();
		QPErrorMsg      SolveKKT ();
		QPErrorMsg      AddVariableToBase();
		QPErrorMsg      AddVariableToBase( Index  i, TypeEvent WhereNext );
		QPErrorMsg      FromBaseToZero ( Index i );
		QPErrorMsg      FromBaseToC ( Index i );
		QPErrorMsg      FromCToBase ( Index i );
		QPErrorMsg      FromZeroToBase ( Index i );
		QPErrorMsg      MoveAlongDirection ();
		QPErrorMsg      RecomputeAll ();
//---------------------------------------------------------------------------
		FILE            *out_info;

		Index			PrintMode;
//----------------------------------------------------------------------------
//      Computation of Regularized Path
//----------------------------------------------------------------------------
		void            ComputeMinRatioTestStep1();
		inline hpReal   GetTheta();
		QPErrorMsg      SetTheta( hpReal val );
        inline hpReal   GetCurrentMinRationStep1Pos();
        inline hpReal   GetCurrentMinRationStep1Neg();
        inline Index    GetNumberOfTiesStep1();
        void            CopyCurrentDirectionSparse( hpVec Vec );
        inline TypeEvent GetParametricEventStep1();
        inline Index    GetParametricIndexStep1();
		inline hpReal   GetNumericDerivate_f ();
        inline boolean  GetInfiniteDirection();
//----------------------------------------------------------------------------
// Error Testing
//-----------------------------------------------------------------------------
        QPErrorMsg      VerifyQPdata ();
        hpReal          VerifyMinReducedCost ();

		QPErrorMsg      CheckCycling ();

		hpReal          f_deriv_num;
private:

		void            MoveIndexTo_Ic  ( Index i, hpReal QicCc, hpReal Qii );
		void            MoveIndexOut_Ic ( Index i , hpReal QicCc, hpReal Qii );
		void            MoveIndexTo_Io  ( Index i );
		void            MoveIndexOut_Io ( Index i );
		void            MoveIndexTo_Is  ( Index i, hpReal linobj, hpReal InitVal );
        void            MoveIndexOut_Is ( Index i );

//-----------------------------------------------------------------------------
// Virtual Functions...
//-----------------------------------------------------------------------------

		// Possible Interface...
        virtual void            UpdateEveryIteration () = 0;

        // Interface with Reduce Cost...
        virtual inline hpReal   GetQicCc ( Index i )=0;
        virtual boolean         NeedToGetNewIndex ()=0;
        virtual void            RedCostIndexAddedToIc ( Index i )=0;
        virtual void            RedCostIndexAddedToIo ( Index i )=0;
		virtual void            RedCostIndexRemovedFromIc ( Index i )=0;
        virtual void            RedCostIndexRemovedFromIo ( Index i )=0;

        // Interface with Matrix Q...
        virtual inline hpReal   GetDataQij ( Index i, Index j ) = 0;
                                // return element Qij from data itself
        virtual inline hpReal   GetQij ( Index i, Index j ) = 0;
                                // return element Qij
        virtual inline hpReal   SprodQi_dIset ( Index i, hpVec dIset, IndexVec Iset, Index Nset)=0;
                                // return < Qi[Iset], dIset[1...Nset] >, NOTE: dIset has size Nset
        virtual inline hpReal   SprodQi_d ( Index i, hpVec d, IndexVec Iset, Index Nset)=0;
								// return < Q[Iset], d[Iset]>, NOTE: d has size n
        virtual QPErrorMsg      FreeMemQ( Index i )=0;
        virtual QPErrorMsg      AddColumnToQ( Index i )=0;
        virtual QPErrorMsg      VerifyQPMem ()=0;
//---------------------------------------------------------------------------
//  Operation over L....
//---------------------------------------------------------------------------
		void            AddRowToLD ( Index i, Index place_i );
        void            AddRowToLI ( Index i, Index place_i, hpReal delta );

        QPErrorMsg      AllocateL( Index i );
        QPErrorMsg      SwapLDMemL ( Index from_i, Index to_i);
        QPErrorMsg      FreeMemL( Index Is_index );
        QPErrorMsg      RemoveRowColumnL( Index i );
        void            DeleteL();
//---------------------------------------------------------------------------
// Indices and Sizes

IntNumber       Iteration;              // Current Number of Iterations Step 1 - Step 2 Loop
IntNumber       InnerIter;              // Current Number of Inner Iterations (KKT)

Index           Max_n;                  // Maximum number of components allocated (obj needs to be destroyed to change)
                                        // might be useful to the incremental mode
Index           n;                      // Actual maximum number of components used (n<= Max_n), can change.

Index           Ns;                     // Number of Active components
Index           Nc;                     // Number of components fixed at c
Index           No;                     // Number of components fixed at 0

IndexVec        Is;                     // Indices of active components
IndexVec        Ic;                     // Indices fixed at c
IndexVec        Io;                     // Indices fixed at 0

IndexVec        PlaceIs;                // PlaceIs[i]= position on Is of index i,
IndexVec        PlaceIc;                //             (Max_n if i not in Is)
IndexVec        PlaceIo;                // similar...

//---------------------------------------------------------------------------
//  Original QP data storage

hpVec   Y;                              // linear equality constraint
hpVec   C;                              // box constraint
hpVec   V;                              // linear term of the objective function

boolean NoBias;                         // == 0 if the bias b is allowed to be nonzero
                                        // == 1 if bias b = 0 (which eliminates the equality constraint) 
//---------------------------------------------------------------------------
//  Working data to solve QP

hpMat   L;                              // Cholesky decomposition, L L^T = Qss

hpVec   r1;                             // Auxiliary vector r1 = L^{-1} Ys
hpVec   r2;                             // Auxiliary vector r2 = L^{-1}( Qsc Cc - Vs )

hpReal  b;                              // Lagrange Multiplier b = (<r1,r2> - <Yc,Cc>)/<r1,r1>

hpVec   alpha;                          // spase vector of alpha, only ns = |Is| components
                                        // alpha_s = L^{-T} ( b r1 - r2 )  (in Case 0)

hpVec   lin_obj;                        // linear term of the objective function, |Is| components
hpReal  cte_obj;                        // constant term of the objective function
hpReal  f_obj;                          // current objective solution

hpReal  r1r1;                           // r1r1 = ||r1||^2
hpReal  r1r2;                           // r1r2 = <r1,r2>
hpReal  r1Lh1;                          // r1Lh1 = <r1,Lh1>  Lh1 is the first linear dependent row
hpReal  r2Lh1;                          // r2Lh1 = <r2,Lh1>
hpReal  r1Lh2;                          // r1Lh2 = <r2,Lh1>  Lh2 is the last linear dependent row
hpReal  r2Lh2;                          // r2Lh2 = <r2,Lh2>

hpReal  YcCc;                           // YcCc  = <Yc,Cc>

hpReal  b_iter;                         // current b
hpVec   alpha_iter;                     // current alpha
hpVec   dd;                             // descend direction from alpha_iter

hpVec   tmpVec;                         // Temporary Vector of size 2*Max_n

                                        // Machine Error...
hpReal  AbsTolError;                    // Absolute Precision
hpReal  RelTolError;                    // Relative Precision

//---------------------------------------------------------------------------
// Data Used to Avoid Cycles
//---------------------------------------------------------------------------

IntNumber MaxIteration;
Index OldIndex;
hpReal Old_f_obj;

//---------------------------------------------------------------------------
// Data associated with Regularized Path
//---------------------------------------------------------------------------

hpVec   z1;                             // r2 = theta z1 - z2
hpVec   z2;

hpReal  d_norm1;                        // || d ||_1
hpReal  r1z1;                           // < r1 , z1 >
hpReal  r1z2;                           // < r1 , z2 >
hpReal  b_deriv;                        // derivative of b w.r.t. theta

hpReal  theta;                          // current regularization parameter
Index   NumberOfTiesStep1;              // = 0 only 1 binding index on the break point
hpReal  MinRatioStep1Pos;               // value of the MinRatio test for Step 1 Increasing theta (Variation of delta)
hpReal  MinRatioStep1Neg;               // value of the MinRatio test for Step 1 Decreasing theta (Variation of delta)
boolean DataConsistentTheta;            // true if current partition if optimal for theta
boolean InfiniteDirection;

hpReal  lin_cte_obj;                    // cte_obj linear in theta
hpReal  quad_cte_obj;                   // cte_obj quadratic in theta

//---------------------------------------------------------------------------
// Flags and Indicators
//---------------------------------------------------------------------------

QPErrorMsg      QPStatus;               // Keep Track of Running Status of th QP
Index           CountMsgInvalidIndex;   // Keep track of Number of inconsistencies of NeedToGetNewIndex
Index           CountMsgNumericProblem; // Keep track of Number of Recomputations from the scratch

Index           DegreeDependence;       // Number of Dependent Rows in Q (0, 1, 2)

TypeEvent       BaseEvent;              // Next base management required to be made.

boolean         LocalOptima;            // = 1 if alpha_iter Current point is a Local, =0 otherwise
Index           BindingIndex;           // Index that achieves the Min-ratio test
hpReal          StepSize;               // Length of the Step Size

Index           NewIndexToBeAdded;      // Index returned by NeedToGetNewIndex()
hpReal          ReducedCostMin;         // Associated Reduced Cost

IntNumber       InfiniteDirectionCount; // Count the number of Infinite Directions...
IntNumber       Case0Count;             // Count the number of Case 0 in KKT...
IntNumber       Case1Count;             // Count the number of Case 1 in KKT...
IntNumber       Case2Count;             // Count the number of Case 2 in KKT...

//----------------------------------------------------------------------------
// Memmory Allocation Parameters
//----------------------------------------------------------------------------

IntNumber       BlockAllocL;             // Number of Rows allocatted at once (we must have BlockAlloc < Max_n)
IntNumber       CurrentAllocL;          // Current Number of allocated rows

}; //----- end of TQPclassPath declaration ----


//-----------------------------------------------------------------------------
// Inline functions
//-----------------------------------------------------------------------------
inline	void	TQPclassPath::SetPrintMode( IntNumber val )
{
PrintMode = val;
}
//-----------------------------------------------------------------------------
inline void 	TQPclassPath::SetBlockAllocL ( IntNumber val )
{
BlockAllocL = val;
}
//-----------------------------------------------------------------------------
inline Index TQPclassPath::Getn()
{
return n;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPclassPath::Getf_obj()
{
return f_obj;
}
//-----------------------------------------------------------------------------
inline Index TQPclassPath::GetCurrentAllocL()
{
return CurrentAllocL;
}
//-----------------------------------------------------------------------------
inline Index TQPclassPath::GetDegreeDependence()
{
return DegreeDependence;
}
//-----------------------------------------------------------------------------
inline hpReal TQPclassPath::GetStepSize()
{
return StepSize;
}
//-----------------------------------------------------------------------------
inline Index TQPclassPath::GetNewIndexToBeAdded()
{
return NewIndexToBeAdded;
}
//-----------------------------------------------------------------------------
inline hpReal TQPclassPath::GetMinReducedCost()
{
return ReducedCostMin;
}
//-----------------------------------------------------------------------------
inline IntNumber TQPclassPath::GetCase0Count()
{
return Case0Count;
}
//-----------------------------------------------------------------------------
inline IntNumber TQPclassPath::GetCase1Count()
{
return Case1Count;
}
//-----------------------------------------------------------------------------
inline IntNumber TQPclassPath::GetCase2Count()
{
return Case2Count;
}
//-----------------------------------------------------------------------------
inline IntNumber TQPclassPath::GetInfiniteDirectionCount()
{
return InfiniteDirectionCount;
}
//-----------------------------------------------------------------------------
inline boolean TQPclassPath::GetInfiniteDirection()
{
return InfiniteDirection;
}
//-----------------------------------------------------------------------------
inline IntNumber TQPclassPath::GetIteration()
{
return Iteration;
}
//-----------------------------------------------------------------------------
inline IntNumber TQPclassPath::GetInnerIter()
{
return InnerIter;
}
//-----------------------------------------------------------------------------
inline void TQPclassPath::SetMaxIteration ( IntNumber val )
{
MaxIteration = val;
}

//-----------------------------------------------------------------------------
inline void     TQPclassPath::CopyCurrentIs( IndexVec Vec )
{
AssignVectors( Vec, Is, Ns );
}
//-----------------------------------------------------------------------------
inline void     TQPclassPath::CopyCurrentDirectionSparse( hpVec Vec )
{
AssignVectors( Vec, dd, Ns );
}
//-----------------------------------------------------------------------------
inline void     TQPclassPath::CopyCurrentIo( IndexVec Vec )
{
AssignVectors( Vec, Io, No );
}
//-----------------------------------------------------------------------------
inline void     TQPclassPath::CopyCurrentIc( IndexVec Vec )
{
AssignVectors( Vec, Ic, Nc );
}
//----------------------------------------------------------------------------
inline Index    TQPclassPath::GetNs()
{
return Ns;
}
//-----------------------------------------------------------------------------
inline Index    TQPclassPath::GetNo()
{
return No;
}
//-----------------------------------------------------------------------------
inline Index    TQPclassPath::GetNc()
{
return Nc;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetAlpha( Index k )
{
return alpha_iter[k];
}
//-----------------------------------------------------------------------------
inline Index    TQPclassPath::GetIs( Index k )
{
return Is[k];
}
//-----------------------------------------------------------------------------
inline Index    TQPclassPath::GetIo( Index k )
{
return Io[k];
}
//-----------------------------------------------------------------------------
inline Index    TQPclassPath::GetIc( Index k )
{
return Ic[k];
}
//-----------------------------------------------------------------------------
inline Index    TQPclassPath::GetMax_n()
{
return Max_n;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPclassPath::Getb()
{
return b_iter;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetYi(Index i)
{
return Y[i];
}
//-----------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetVi(Index i)
{
return V[i];
}
//-----------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetCi(Index i)
{
return C[i];
}
//-----------------------------------------------------------------------------
inline void     TQPclassPath::SetBaseEvent ( TypeEvent CurrentEvent )
{
BaseEvent = CurrentEvent;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetAbsTolError ()
{
return AbsTolError;
}
//-----------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetRelTolError ()
{
return RelTolError;
}
//-----------------------------------------------------------------------------
inline void     TQPclassPath::CopyCurrentPointSparse( hpVec Vec )
{
AssignVectors( Vec, alpha_iter, Ns );
}
//-----------------------------------------------------------------------------

inline void     TQPclassPath::SetNewIndexToBeAdded ( Index i , hpReal val )
{
NewIndexToBeAdded = i;
ReducedCostMin    = val;
}

//-----------------------------------------------------------------------------
#if ( TimeMode )
inline void     TQPclassPath::TimeStart()
{
  //ftime(&TimeStartCount);
TimeStartCount=QPCpuTime();
}

inline void     TQPclassPath::TimeFinish()
{
  //ftime(&TimeFinishCount);
TimeFinishCount=QPCpuTime();
}

inline void     TQPclassPath::TimeAdd( hpReal *time_total )
{
*time_total += (TimeFinishCount.time - TimeStartCount.time) + 1.0e-3 * (TimeFinishCount.millitm - TimeStartCount.millitm) ;
}
//---------------------------------------------------------------------------
inline void     TQPclassPath::TimeTotal()
{
struct timeb temp;
//ftime(&temp);
temp=QPCpuTime();
TimeTotalAll = (temp.time - TimeStartAll.time) + 1.0e-3 * (temp.millitm - TimeStartAll.millitm) ;
}
//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetTotalTime ()
{
return TimeTotalAll;
}
//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetQPTime ()
{
return TimeTotalAll - TimeInit;
}
#endif

//---------------------------------------------------------------------------
inline void     TQPclassPath::SetC ( hpVec C0 )
{
if ( Iteration ) {
		PrintErrorMsg ("Object already in use, cannot update vector C now.\n Create new object.\n");
		}
else    {
		AssignVectors( C, C0, n );
		}
}

//---------------------------------------------------------------------------
inline void     TQPclassPath::SetC ( hpReal C0 )
{
if ( Iteration ) {
		PrintErrorMsg ("Object already in use, cannot update vector C now.\n Create new object.\n");
		}
else    {
		VecInit( C, C0, n );
		}
}

//---------------------------------------------------------------------------
inline void     TQPclassPath::SetAbsPrecision( hpReal val )
{
if ( val > hpZero )     AbsTolError = val;
}
//---------------------------------------------------------------------------
inline void     TQPclassPath::SetRelPrecision( hpReal val )
{
if ( val > hpZero )     RelTolError = val;
}

//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetTheta()
{
return theta;
}
//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::Getb_deriv()
{
return b_deriv;
}
//---------------------------------------------------------------------------
inline Index    TQPclassPath::GetNumberOfTiesStep1()
{
return NumberOfTiesStep1;
}
//---------------------------------------------------------------------------
inline TypeEvent TQPclassPath::GetParametricEventStep1()
{
return BaseEvent;
}
//---------------------------------------------------------------------------
inline Index    TQPclassPath::GetParametricIndexStep1()
{
return BindingIndex;
}
//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetCurrentMinRationStep1Pos()
{
if ( DataConsistentTheta ) return MinRatioStep1Pos;
else    {
		PrintErrorMsg("Data is not consistent with current value of the parameter.");
		return -1;
		}
}
//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetCurrentMinRationStep1Neg()
{
if ( DataConsistentTheta ) return MinRatioStep1Neg;
else    {
		PrintErrorMsg("Data is not consistent with current value of the parameter.");
		return -1;
		}
}
//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::GetNumericDerivate_f ()
{
return f_deriv_num;
}

//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::Getr1r1 ()
{
return r1r1;
}

//---------------------------------------------------------------------------
inline hpReal   TQPclassPath::Getr1z1 ()
{
return r1z1;
}
//---------------------------------------------------------------------------
inline bool 	TQPclassPath::BelongToIo ( Index k )
{
return ( PlaceIo[ k ] < Max_n ) ? true : false;
}
//---------------------------------------------------------------------------
inline bool 	TQPclassPath::BelongToIc ( Index k )
{
return ( PlaceIc[ k ] < Max_n ) ? true : false;
}
//---------------------------------------------------------------------------
inline bool 	TQPclassPath::BelongToIs ( Index k )
{
return ( PlaceIs[ k ] < Max_n ) ? true : false;
}
#endif
