/*---------------------------------------------------------------------------
File Name:      VectorOperations.cpp and VectorOperations.h
Author:         Alexandre Belloni
Last Modified:  20-SEP-05

Description:    We will gather some basic routines to deal with simple vector
                operations like scalar product, sum, weigthed sum, and also
                their restrictions via index sets
-----------------------------------------------------------------------------*/

//---------------------------------------------------------------------------
#ifndef VectorOperationsH
#define VectorOperationsH
//---------------------------------------------------------------------------
// Include Files
//---------------------------------------------------------------------------

#include "BasicInit.h"

//---------------------------------------------------------------------------
// Scalar functions
//---------------------------------------------------------------------------
template<class Tclass>
inline Tclass min( const Tclass x, const Tclass y )
{
return ( x > y ? y : x );
}

//---------------------------------------------------------------------------
template<class Tclass>
inline Tclass max( const Tclass x, const Tclass y )
{
return ( x > y ? x : y );
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void swap( Tclass &x, Tclass &y )
{
const Tclass tmp = x;
x = y;
y = tmp;
}

//---------------------------------------------------------------------------
template<class Tclass>
inline Tclass abs( const Tclass x )
{
return ( x >= 0 ? x : -x );
}


//---------------------------------------------------------------------------
//  Vector Operations
//---------------------------------------------------------------------------

// sprod: Scalar Product between two vector; we demand both objects to be of the same type.
template<class Tclass>
inline Tclass sprod ( const Tclass *x, const Tclass *y, Index n)
{
register Tclass tmp=0;

for ( ; n-- ; )
        tmp += (*(x++))  *  (*(y++));
return tmp;
}

//---------------------------------------------------------------------------
template<class Tclass>
inline Tclass sprod ( const Tclass *x, const Tclass *y, IndexVec IndSet, Index n )
{
register Tclass tmp=0;

for ( register Index i ; n-- ;  ){
        i=IndSet[n];
        tmp += x[i]  *  y[i];
        }
return tmp;
}
//---------------------------------------------------------------------------

template<class Tclass>
inline Tclass norm ( const Tclass *x,  Index n)
{
register Tclass tmp=0;

for ( ; n-- ; ){
        register const Tclass x_i=(*(x++));
        tmp += x_i * x_i;
        }
return tmp;
}

//---------------------------------------------------------------------------
template<class Tclass>
inline Tclass norm ( const Tclass *x, IndexVec IndSet, Index n )
{
register Tclass tmp=0;

for ( register Index i ; n-- , i=IndSet[n] ; ){
        register const Tclass x_i =x[i];
        tmp += x_i  *  x_i;
        }
return tmp;
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void RandomizeVector ( Tclass *vec, long size )
{
long i;
long tmp;
for ( long k = size-1 ; k ; k-- ){
        i = rand() % k;
        tmp = vec[i];
        vec[i] = vec[k];
        vec[k] = tmp;
        }
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template<class Tclass>
inline void SwapVectors( Tclass *x, Tclass *y, Index n )
{
register Tclass tmp;
for ( ; n-- ; ){
        tmp = *x;
        *(x++) = *y;
        *(y++) = tmp;
        }
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void SumVectors ( Tclass *out, const Tclass *x, const Tclass *y,  Index n )
{
for ( ; n-- ; )
        *(out++) = *(x++) +  *(y++);
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void SumVectors ( Tclass *out, const Tclass *x, const Tclass *y, IndexVec IndSet, Index n )
{
for ( register Index i ; n-- , i=IndSet[n] ; )
        out[i] = x[i]  +  y[i];
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void DiffVectors ( Tclass *out, const Tclass *x, const Tclass *y,  Index n )
{
for ( ; n-- ; )
        *(out++) = *(x++) -  *(y++);
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void DiffVectors ( Tclass *out, const Tclass *x, const Tclass *y, IndexVec IndSet, Index n )
{
for ( register Index i ; n-- , i=IndSet[n] ; )
        out[i] = x[i]  -  y[i];
}
//---------------------------------------------------------------------------
template<class Tclass>
inline void ScalingVector ( Tclass *out, const Tclass a, const Tclass *x,  Index n )
{
for ( ; n-- ; )
        *(out++) = a * (*(x++));
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void ScalingVector ( Tclass *out, const Tclass a, const Tclass *x, IndexVec IndSet, Index n )
{
for ( register Index i ; n-- , i=IndSet[n] ; )
        out[i] =a*x[i];
}

//---------------------------------------------------------------------------
template<class Tclass, class TTclass>
inline void WSumVectors ( Tclass *out, const TTclass a, const Tclass *x, const TTclass b, const Tclass *y,  Index n )
{
for ( ; n-- ; )
        *(out++) = a*(*(x++)) + b*(*(y++));
}

//---------------------------------------------------------------------------
template<class Tclass,class TTclass>
inline void WSumVectors ( Tclass *out, const TTclass a, const Tclass *x, const TTclass b, const Tclass *y, IndexVec IndSet, Index n )
{
for ( register Index i ; n-- , i=IndSet[n] ; )
        out[i] = a * x[i]  + b * y[i];
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void VecInit ( Tclass *x, register const Tclass a, Index n )
{
for (register Tclass *aux = x + n ; aux > x ; )
        *(--aux) = a;
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void AssignVectors( Tclass *x, Tclass *y, Index n )
{
for ( ; n-- ; )
        *(x++) = *(y++);
}

//---------------------------------------------------------------------------
template<class Tclass>
inline void ShiftLeft ( Tclass *x, Index Start, Index Final )
{
register Tclass *aux = x + Start;
for (register Index k = Final - Start ; k-- ; )
        *(aux++) = *(aux+1);
}


#endif
